/*
 *
 *  Copyright (c) 2004-2005  Warren Jasper <wjasper@tx.ncsu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "pmd.h"
#include "usb-1208LS.h"

enum Mode {Differential, SingleEnded};

/* configures digital port */
void usbDConfigPort_USB1208LS(HIDInterface* hid, __u8 port, __u8 direction)
{

  struct {
    __u8 cmd;
    __u8 port;
    __u8 direction;
    __u8 pad[5];
  } report;

  //  report.report_id = 0;
  report.cmd = DCONFIG;
  report.port = port;
  report.direction = direction;

  PMD_SendOutputReport(hid, 0, (__u8*) &report, sizeof(report), LS_DELAY);
}

/* reads digital port  */
void usbDIn_USB1208LS(HIDInterface* hid, __u8 port, __u8* din_value)
{
  __u8 cmd[8];
  
  cmd[0] = DIN;
  cmd[1] = port;
  
  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, din_value, 1, LS_DELAY);
}

/* writes digital port */
void usbDOut_USB1208LS(HIDInterface* hid, __u8 port, __u8 value) 
{
  __u8 cmd[8];
  
  cmd[0] = DOUT;
  cmd[1] = port;
  cmd[2] = value;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

/* reads digital port bit */
__u8 usbDBitIn_USB1208LS(HIDInterface* hid, __u8 port, __u8 bit) 
{
  __u8 cmd[8];
  __u8 value;

  cmd[0] = DBIT_IN;
  cmd[1] = port;
  cmd[2] = bit;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, &value, sizeof(value), LS_DELAY);

  return value;
}

/* writes digital port bit */
void usbDBitOut_USB1208LS(HIDInterface* hid, __u8 port, __u8 bit, __u8 value)
{
  __u8 cmd[8];
  
  cmd[0] = DBIT_OUT;
  cmd[1] = port;
  cmd[2] = bit;
  cmd[3] = value;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

/* reads from analog in */
signed short usbAIn_USB1208LS(HIDInterface* hid, __u8 channel, __u8 range)
{
  enum Mode mode;

  __u8 cmd[8];
  __s16 value;

  cmd[0] = AIN;
  cmd[1] = channel;
  cmd[2] = range;

  if ( range == SE_10_00V ) {
    mode = SingleEnded;
  } else {
    mode = Differential;
  }

  if (channel > 3 && mode == Differential ) {
    printf("usbAIN: channel out of range for differential mode.\n");
    return -1;
  }

  if (channel > 7 && mode == SingleEnded ) {
    printf("usbAIN: channel out of range for single ended mode.\n");
    return -1;
  }

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, (__u8 *) &value, sizeof(value), LS_DELAY);
  if ( mode == Differential ) {
    /* the data is a 2's compliment signed 12 bit number */
    value /= (1<<4);
  } else {
    /* the data is a  2's compliment signed 11 bit number */
    if ( value > 0x7ff0 ) {
      value = 0;
    } else if ( value > 0x7fe0 ) {
      value = 0xfff;
    } else {
      value >>= 3;
      value &= 0xfff;
    }
    value -= 0x800;
  }
  return value;
}

  
/* writes to analog out */
void usbAOut_USB1208LS(HIDInterface* hid, __u8 channel, __u16 value) 
{
  __u8 cmd[8];

  cmd[0] = AOUT;
  cmd[1] = channel;
  cmd[2] = (__u8) (value & 0xff);
  cmd[3] = (__u8) ((value >> 8) & 0xff);
  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

void usbAInScan_USB1208LS(HIDInterface* hid, __u16 count, int rate, __u8 low_channel, __u8 high_channel, __u8 options, __s16 value[], __u8 gainLoadQueue[])
{

  int i, idx;
  int scans;
  __u16 scan_index;
  __u16 actual_scan_index;
  __u8 chanCount;
  __u8 chanLoadQueue[8];
  
  struct {
    __u8 cmd;
    __u8 lo_count;
    __u8 hi_count;
    __u8 timer_preload;
    __u8 timer_prescale;
    __u8 options;
    __u8 extra[2];
  } out;

  struct t_Feature_Report{
  //    __u8 reportID;
    __u8 data[96];
    __u8 error;
    __u8 readAddress[2];
    __u8 writeAddress[2];
    __u8 scanIndex[2];
    __u8 extra;
  } feature_report;
  
  __u8 buffer[9] = { 0, 0, 0, 0, 0, 0, 0, 0};

  int timerMult;
  __u8 timerVal;
  __u8 prescale;
  __u8 preload;
  __u8 setupTime;

  usbAInStop_USB1208LS(hid);   // just to make sure.

  if ((100 <= rate) && (rate < 200)) {       // Select 256:1 prescaler
    prescale = 7;
    setupTime = 0;
  } else if ((200 <= rate) && (rate < 400)) { // Select 128:1 prescaler
    prescale = 6;
    setupTime = 0;
  } else if ((400 <= rate) && (rate < 800)) { // Select 64:1 prescaler
    prescale = 5;
    setupTime = 0;
  } else if ((800 <= rate) && (rate < 1500)) { // Select 32:1 prescaler
    prescale = 4;
    setupTime = 1;
  } else if ((1500 <= rate) && (rate < 3000)) { // Select 16:1 prescaler
    prescale = 3;
    setupTime = 3;
  } else if ((3000 <= rate) && (rate < 6000)) { // Select 8:1 prescaler
    prescale = 2;
    setupTime = 6;
  } else if ((6000 <= rate) && (rate < 8192)) { // Select 4:1 prescaler
    prescale = 1;
    setupTime = 10;
  } else {
    printf("usbAInScan_USB1208LS: sampling rate out of range.\n");
    return;
  }

  timerMult = 1 << (prescale + 1);
  timerVal = (__u8) ((256 - (MINILAB_CLOCK / (rate * timerMult))) + 0.5);
  preload = timerVal + setupTime;

  /* set up gain queue */
  chanCount = high_channel - low_channel + 1;
  for ( i = 0; i < chanCount; i++ ) {
    chanLoadQueue[i] = low_channel + i;
  }
  usbAInLoadQueue_USB1208LS(hid, chanCount, chanLoadQueue, gainLoadQueue);

  out.cmd = AIN_SCAN;
  out.lo_count = count & 0xff;
  out.hi_count = (count >> 8) & 0xff;
  out.timer_preload = preload;
  out.timer_prescale = prescale;
  out.options = options;

  PMD_SendOutputReport(hid, 0, (__u8*) &out, sizeof(out), LS_DELAY);

  /*
     If in external trigger mode, then wait for the device to send back
     notice that the trigger has been received, then startup the acquisition
  */

  if ( options & AIN_TRIGGER ) {
    while ( buffer[1] != 0xC3 ) {  // wait until external trigger received
      buffer[0] = 0;
      PMD_GetInputReport(hid, 0, (__u8  *) &buffer, sizeof(buffer), LS_DELAY*10);
    }
  }

  /*
    In Burst Mode, wait for End of Block acquisition flag (0xA5)
  */

  if ( options & AIN_BURST_MODE ) {
    buffer[1] = 0;
    while ( buffer[1] != 0xA5 ) {  // wait until external trigger received
      buffer[0] = 0;
      PMD_GetInputReport(hid, 0, (__u8  *) &buffer, sizeof(buffer), LS_DELAY);
    }
  }
  
  /*
    Retrieve the AInScan Response
    GET HID Feature Report to collect the data buffer.  Each buffer will be 105
    bytes long.  The first bye will contain the record number and can be ignored.
    The following 96 bytes will represent 64 samples of data.
  */

  feature_report.scanIndex[0] = 0xff;
  feature_report.scanIndex[1] = 0xff;
  scan_index = 0;
  idx = 0;
  for ( scans = 0; scans < count/65 + 1; scans++ ) {
    memset(&feature_report, 0xbeef, sizeof(feature_report));
    do {
      PMD_GetFeatureReport(hid, 0, (__u8 *) &feature_report, sizeof(feature_report), 2000);
      actual_scan_index = (__u16) (feature_report.scanIndex[0] | feature_report.scanIndex[1] << 8);
    } while ( scan_index != actual_scan_index );
    scan_index++;
    printf("Completed scan %d  error = %d\n", scan_index, feature_report.error);
    for ( i = 0; i < 96; i += 3, idx += 2 ) {
      printf("data[%d] = %#x  data[%d] = %#x\n", i, feature_report.data[i], i+1, feature_report.data[i+1]);
      value[idx] = feature_report.data[i] | ((feature_report.data[i+1]<<4) & 0x0f00);
      value[idx + 1] = feature_report.data[i+2] | ((feature_report.data[i+1]<< 8) & 0x0f00);
    }
  }
}

void usbAInLoadQueue_USB1208LS(HIDInterface* hid, __u8 chanCount, __u8 chanLoadQueue[], __u8 gainLoadQueue[])
{
  int i;   
  struct {
    __u8 cmd;
    __u8 count;
    __u8 gains[6];
  } out;

  out.cmd = ALOAD_QUEUE;
  out.count = chanCount;   // can be 1, 2, 4, or 8
  chanCount = (chanCount == 8) ? 6 : chanCount;
  for ( i = 0; i < chanCount; i++ ) {
    out.gains[i] = (chanLoadQueue[i] & 0x7) | gainLoadQueue[i] | 0x80;
  }
  PMD_SendOutputReport(hid, 0, (__u8*) &out, sizeof(out), LS_DELAY);

  // Configure the rest of the channels (channel 6 and 7 )
  if ( out.count == 8 ) {
    out.count = 0x2;
    for ( i = 6; i < 8; i++ ) {
    out.gains[6-i] = (chanLoadQueue[i] & 0x7) | gainLoadQueue[i] | 0x80;  
    }
    PMD_SendOutputReport(hid, 0, (__u8*) &out, sizeof(out), LS_DELAY);
  }
}

void usbAInStop_USB1208LS(HIDInterface* hid)
{
  __u8 cmd[8];
  
  cmd[0] = AIN_STOP;
  
  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}  


/* Initialize the counter */
void usbInitCounter_USB1208LS(HIDInterface* hid)
{
  __u8 cmd[8];
  
  cmd[0] = CINIT;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

__u32 usbReadCounter_USB1208LS(HIDInterface* hid)
{
  __u8 cmd[8];
  __u32 value;

  cmd[0] = CIN;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, (__u8  *) &value, sizeof(value), LS_DELAY);

  return value;
}

void usbReadMemory_USB1208LS(HIDInterface* hid, __u16 address, __u8 *data, __u8 count)
{
  __u8 cmd[8];
 
  cmd[0] = MEM_READ;
  cmd[1] = (__u8) (address & 0xff);  // low byte
  cmd[2] = (__u8) (address >> 0x8);  // high byte
  cmd[3] = count;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, data, count, LS_DELAY);
}

/* blinks the LED of USB device */
void usbBlink_USB1208LS(HIDInterface* hid)
{
  struct {
    __u8 cmd;
    __u8 pad[7];
  } report;

  report.cmd = BLINK_LED;
  PMD_SendOutputReport(hid, 0, (__u8*) &report, sizeof(report), LS_DELAY);
}

/* resets the USB device */
int usbReset_USB1208LS(HIDInterface* hid)
{
  __u8 cmd[8];

  cmd[0] = RESET;

  return PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

/* configure external triger */
void usbSetTrigger_USB1208LS(HIDInterface* hid, __u8 type, __u8 chan)
{
  __u8 cmd[8];

  cmd[0] = SET_TRIGGER;
  cmd[1] = type;
  cmd[2] = chan;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

__u8 usbGetID_USB1208LS(HIDInterface* hid)
{
  __u8 cmd[8];
  __u8 data;

  cmd[0] = GET_ID;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
  PMD_GetInputReport(hid, 0, &data, 1, LS_DELAY);

  return data;
}

void usbSetID_USB1208LS(HIDInterface* hid, __u8 id)
{
  __u8 cmd[8];

  cmd[0] = SET_ID;
  cmd[1] = id;

  PMD_SendOutputReport(hid, 0, cmd, sizeof(cmd), LS_DELAY);
}

float volts_LS( const int gain, const signed short num )
{
  float volt;

  switch( gain ) {
    case SE_10_00V:
      volt = num * 10.0 / 0x7ff;
      break;
    case BP_20_00V:
      volt = num * 20.0 / 0x7ff;
      break;
    case BP_10_00V:
      volt = num * 10.0 / 0x7ff;
      break;
    case BP_5_00V:
      volt = num * 5.0 / 0x7ff;
      break;
    case BP_4_00V:
      volt = num * 4.0 / 0x7ff;
      break;
    case BP_2_50V:
      volt = num * 2.5 / 0x7ff;
      break;
    case BP_2_00V:
      volt = num * 2.0 / 0x7ff;
      break;
    case BP_1_25V:
      volt = num * 1.25 / 0x7ff;
      break;
    case BP_1_00V:
      volt = num * 1.0 / 0x7ff;
      break;
  }
  return volt;
}
